#include <algorithm>
#include <cmath>

#include "draw.h"
#include "image.h"

struct vec2
{
    double x, y;
};

struct ray
{
    // origin
    vec2 ori;

    // normalized direction
    vec2 dir;
};

struct circle
{
    // center
    vec2 c;

    // radius
    double r;
};

double intersect(ray r, circle c)
{
    r.ori.x -= c.c.x;
    r.ori.y -= c.c.y;

    double A = r.dir.x * r.dir.x + r.dir.y * r.dir.y;
    double B = (r.ori.x * r.dir.x + r.ori.y * r.dir.y) * 2.0f;
    double C = (r.ori.x * r.ori.x + r.ori.y * r.ori.y) - c.r * c.r;

    double d = B * B - 4.0f * A * C;

    if (d < 0.0f)
    {
        // No hit
        return -1.0f;
    }

    double dd = B < 0.0f ? std::sqrt(d) : -std::sqrt(d);

    double q = -0.5f * (B + dd);

    double t1 = q / A;
    double t2 = C / q;

    if (t1 < 0)
    {
        return t2;
    }

    if (t2 < 0)
    {
        return t1;
    }

    return std::min(t1, t2);
}

vec2 reflect(vec2 wi, vec2 n)
{
}

vec2 refract(vec2 wi, vec2 n, double eta)
{
}

int main()
{
    image_rgb8 img(800, 600);

    std::fill(img.data(), img.data() + img.width() * img.height(), RGB{ 255, 255, 255 });

    // Index of refraction outside circle
    double eta1 = 1.3;
    // Index of refraction inside circle
    double eta2 = 1.0;

    circle c = { 400, 300, 100 };
    draw_circle(c.c.x, c.c.y, c.r, img);

    // Primary rays
    for (int y = 205; y < 400; y += 5)
    {
        ray r = { 50, static_cast<double>(y), 1, 0 };
        double t = intersect(r, c);

        // Point of intersection
        double px = r.ori.x + r.dir.x * t;
        double py = r.ori.y + r.dir.y * t;

        draw_line(r.ori.x, r.ori.y, px, py, img);

        // Calculate normal vector at hit point
        double nx = (px - c.c.x) / c.r;
        double ny = (py - c.c.y) / c.r;

        // Viewing direction
        double vx = -r.dir.x;
        double vy = -r.dir.y;

        vec2 dir = refract(vec2{ vx, vy }, vec2{ nx, ny }, eta1 / eta2);

        // Handle TIR
        // On TIR, refract() should return { 0.0, 0.0 }
        if (std::abs(dir.x) < 0.001 && std::abs(dir.y) < 0.001)
        {
            dir = reflect(vec2{ vx, vy }, vec2{ nx, ny });
        }

        // Normalize
        double len = std::sqrt(dir.x * dir.x + dir.y * dir.y);
        dir.x /= len;
        dir.y /= len;

        r.ori = { px + 0.001 * dir.x, py + 0.001 * dir.y };
        r.dir = dir;

        t = intersect(r, c);

        if (t < 0.0)
        {
            // If we do not hit anything, draw a 200px wide line
            t = 200.0;
        }

        // Point of intersection
        px = r.ori.x + r.dir.x * t;
        py = r.ori.y + r.dir.y * t;

        draw_line(r.ori.x, r.ori.y, px, py, img);

    }

    save_image("out.pnm", img);
}
